package ij.io;
import java.awt.*;
import java.awt.image.*;
import java.io.*;
import ij.*;
import ij.gui.*;
import ij.process.*;

/** Opens tiff (and tiff stacks), gif, jpeg or bmp images,
	and look-up tables, using a file open dialog or a path. */
public class Opener {

	private static final int UNKNOWN=0, DICOM=2, FITS=3, PGM=4;
	private static String defaultDirectory = null;

	public Opener() {
	}

	/** Displays a file open dialog box and then opens,
	the tiff, gif, jpeg, lut or bmp file specified by the user. */
	public void openImage() {
		OpenDialog od = new OpenDialog("Open...", "");
		String directory = od.getDirectory();
		String name = od.getFileName();
		if (name==null)
			return;
		IJ.showStatus("Opening: " + directory + name);
		ImagePlus imp = openImage(directory, name);
		if (imp!=null) imp.show();
	}

	/** Attempts to open the specified file as a tiff, bmp,
	gif, jpeg, bmp or lut. Returns an ImagePlus object if successful. */
	public ImagePlus openImage(String directory, String name) {
		// try to open as tiff
		ImagePlus imp = openTiff(directory, name);
		if (imp!=null)
			return imp;

		String path = directory+name;
		switch (getFileType(path)) {
		case DICOM:
			imp = (ImagePlus)IJ.runPlugIn("ij.plugin.DICOM", path);
			if (imp.getWidth()!=0) return imp; else return null;
		case FITS:
			imp = (ImagePlus)IJ.runPlugIn("ij.plugin.FITS", path);
			if (imp.getWidth()!=0) return imp; else return null;
		case PGM:
			imp = (ImagePlus)IJ.runPlugIn("ij.plugin.PGM_Reader", path);
			if (imp.getWidth()!=0) return imp; else return null;
		}
		
		if (name.endsWith(".lut")) {
			imp = (ImagePlus)IJ.runPlugIn("ij.plugin.LutLoader", path);
			if (imp.getWidth()!=0) return imp; else return null;
		}
		
		if (name.endsWith(".bmp")) {
			imp = (ImagePlus)IJ.runPlugIn("ij.plugin.BMP", path);
			if (imp.getWidth()!=0) return imp; else return null;
		}
				
		// if not tiff, lut or bmp, try to open as gif or jpg
		Image img = Toolkit.getDefaultToolkit().getImage(path);
    	if (img!=null) {
	    	imp = new ImagePlus(name, img);
	    	if (imp.getType()==ImagePlus.COLOR_RGB)
	    		convertGrayJpegTo8Bits(imp);
	    	FileInfo fi = new FileInfo();
	    	fi.fileFormat = fi.GIF_OR_JPG;
	    	fi.fileName = name;
	    	fi.directory = directory;
	    	imp.setFileInfo(fi);
	    	return imp;
	    } else {
IJ.write("Raw");
			imp = (ImagePlus)IJ.runPlugIn("ij.plugin.Raw", path);
			if (imp.getWidth()!=0) return imp; else return null;
	    }
	}
	
	/** If this image is grayscale, convert it to 8-bits. */
	public static void convertGrayJpegTo8Bits(ImagePlus imp) {
		ImageProcessor ip = imp.getProcessor();
		int width = ip.getWidth();
		int height = ip.getHeight();
		int[] pixels = (int[])ip.getPixels();
		int c,r,g,b,offset;
		for (int y=0; y<(height-8); y++) {
			offset = y*width;
			for (int x=0; x<(width-8); x++) {
				c = pixels[offset+x];
				r = (c&0xff0000)>>16;
				g = (c&0xff00)>>8;
				b = c&0xff;
				if (!((r==g)&&(g==b))) {
					//IJ.write("count: "+count+" "+r+" "+g+" "+b);
					return;
				}
			}
			//count++;
		}
		IJ.showStatus("Converting to 8-bits");
		new ImageConverter(imp).convertToGray8();
	}

	/** Are all the images in this file the same size and type? */
	boolean allSameSizeAndType(FileInfo[] info) {
		boolean sameSizeAndType = true;
		boolean contiguous = true;
		int startingOffset = info[0].offset;
		int size = info[0].width*info[0].height*info[0].getBytesPerPixel();
		for (int i=1; i<info.length; i++) {
			sameSizeAndType &= info[i].fileType==info[0].fileType
				&& info[i].width==info[0].width
				&& info[i].height==info[0].height;
			contiguous &= info[i].offset==startingOffset+i*size;
		}
		if (contiguous)
			info[0].nImages = info.length;
		if (IJ.debugMode) {
			IJ.write("  sameSizeAndType: " + sameSizeAndType);
			IJ.write("  contiguous: " + contiguous);
		}
		return sameSizeAndType;
	}
	
	/** Attemps to open the file as a stack. Returns 
		an ImagePlus object if successful. */
	public ImagePlus openTiffStack(FileInfo[] info) {
		if (info.length>1 && !allSameSizeAndType(info))
			return null;
		FileInfo fi = info[0];
		if (fi.nImages>1)
			return new FileOpener(fi).open(false); // open contiguous images as stack
		else {
			ColorModel cm = FileOpener.createColorModel(fi);
			ImageStack stack = new ImageStack(fi.width, fi.height, cm);
			Object pixels;
			int skip = fi.offset;
			int imageSize = fi.width*fi.height*fi.getBytesPerPixel();
			int loc = 0;
			try {
				InputStream is = FileOpener.createInputStream(fi);
				ImageReader reader = new ImageReader(fi);
				for (int i=0; i<info.length; i++) {
					IJ.showStatus("Reading: " + (i+1) + "/" + info.length);
					pixels = reader.readPixels(is, skip);
					if (pixels==null) break;
					loc += imageSize+skip;
					if (i<(info.length-1)) {
						skip = info[i+1].offset-loc;
						if (skip<0) throw new IOException("Images are not in order");
					}
					if (fi.fileType==fi.GRAY16_UNSIGNED)
						stack.addUnsignedShortSlice(null, pixels);
					else
						stack.addSlice(null, pixels);
					
					IJ.showProgress((double)i/info.length);
				}
			}
			catch (Exception e) {
				IJ.write("" + e);
			}
			catch(OutOfMemoryError e) {
				IJ.outOfMemory(fi.fileName);
				stack.deleteLastSlice();
				stack.deleteLastSlice();
			}
			IJ.showProgress(1.0);
			if (stack.getSize()==0)
				return null;
			if (fi.whiteIsZero)
				new StackProcessor(stack, stack.getProcessor(1)).invert();
			ImagePlus imp = new ImagePlus(fi.fileName, stack);
			imp.setFileInfo(fi);
			IJ.showProgress(1.0);
			return imp;
		}
	}
	
	/** Attempts to open the specified file as a tiff.
		Returns an ImagePlus object if successful. */
	ImagePlus openTiff(String directory, String name) {
		TiffDecoder td = new TiffDecoder(directory, name);
		if (IJ.debugMode) td.enableDebugging();
		ImagePlus imp=null;
		FileInfo[] info=null;
		try {info = td.getTiffInfo();}
		catch (IOException e) {
			IJ.write("TiffDecoder: " + e.getMessage());
			return null;
		}
		if (info==null)
			return null;
		if (IJ.debugMode) // dump tiff tags
			IJ.write(info[0].info);
		if (info.length>1) { //try to open as stack
			imp = openTiffStack(info);
			if (imp!=null) {
				FileOpener.setResolution(info[0], imp);
				return imp;
			}
		}
		FileOpener fo = new FileOpener(info[0]);
		imp = fo.open(false);
		if (info.length>1)
			IJ.write((info.length-1) + " unopened images in this file");
		IJ.showStatus("");
		return imp;
	}
	
	/**
	Attempts to determinate the image file type by looking for
	'magic numbers' or text strings in the header.
	 */
	int getFileType(String path) {
		InputStream is;
		byte[] buf = new byte[132];
		try {
			is = new FileInputStream(path);
			is.read(buf, 0, 132);
			is.close();
		} catch (IOException e) {
			return UNKNOWN;
		}
		
		 // look DICOM file with "DICM" at offset 128
		if (buf[128]==68 && buf[129]==73 && buf[130]==67 && buf[131]==77)
			return DICOM;

 		// look for ACR/NEMA file
 		if (buf[0]==8 && buf[1]==0) {
 			 if (buf[2]==0 && buf[3]==0 && buf[4]==4) //(00008,0000), length=4
 			 	return DICOM;
 			 if (buf[2]==8 && buf[3]==0 && buf[4]==16) //(00008,0008), length=16
 			 	return DICOM;
 		}

		// look for FITS file starting with "SIMP"
		if (buf[0]==83 && buf[1]==73 && buf[2]==77 && buf[3]==80)
			return FITS;
			
		// look for PGM file starting with "P2" or "P3"
		if (buf[0]==80&&(buf[1]==50||buf[1]==53)&&(buf[2]==10||buf[2]==13))
			return PGM;

		return UNKNOWN;
	}

}
