package ij.process;
import ij.measure.Measurements;
import java.awt.*;

/** Statistics, including the histogram, of an image or image roi. */
public class ImageStatistics implements Measurements {

	public int[] histogram;
	public double min, max;
	public int pixelCount;
	public int mode;
	public double mean;
	public double stdDev;
	public double xCentroid;
	public double yCentroid;
	public double xCenterOfMass;
	public double yCenterOfMass;
	public int roiX, roiY, roiWidth, roiHeight;
	
	public double histMin;
	public double histMax;
	public int maxCount;
	public int nBins;
	public double binSize = 1.0;
	
	protected int width, height;
	
	public static ImageStatistics getStatistics(ImageProcessor ip, int mOptions, int nBins) {
		if (ip instanceof ByteProcessor)
			return new ByteStatistics(ip, mOptions, nBins);
		else if (ip instanceof ShortProcessor)
			return new ShortStatistics(ip, mOptions, nBins);
		else if (ip instanceof ColorProcessor)
			return new ColorStatistics(ip, mOptions, nBins);
		else
			return new FloatStatistics(ip, mOptions, nBins);
	}

	void getMinAndMax() {
		int min = 0;
		while ((histogram[min] == 0) && (min < 255))
			min++;
		this.min = min;
			
		int max = 255;
		while ((histogram[max] == 0) && (max > 0))
			max--;
		this.max = max;
	}

	void getStatistics(int minThreshold, int maxThreshold) {
		int count;
		double value;
		double sum = 0;
		double sum2 = 0;
		
		histMin = 0;
		histMax = 256;
		nBins = 256;
		binSize = (histMax-histMin)/nBins;
		for (int i=minThreshold; i<=maxThreshold; i++) {
			count = histogram[i];
			pixelCount += count;
			value = histMin+i*binSize;
			sum += value*count;
			sum2 += (value*value) * count;
			if (count > maxCount) {
				maxCount = count;
				mode = i;
			}
		}
		mean = sum / pixelCount;
		if (pixelCount>0) {
			stdDev = (pixelCount*sum2-sum*sum)/pixelCount;
			if (stdDev>0.0)
				stdDev = Math.sqrt(stdDev/(pixelCount-1.0));
			else
				stdDev = 0.0;
		} else
			stdDev = 0.0;
		calculateStdDev(pixelCount, sum, sum2);
	}
	
	void calculateStdDev(int n, double sum, double sum2) {
		if (n>0) {
			stdDev = (n*sum2-sum*sum)/n;
			if (stdDev>0.0)
				stdDev = Math.sqrt(stdDev/(n-1.0));
			else
				stdDev = 0.0;
		}
		else
			stdDev = 0.0;
	}
		
	void setRoiLimits(ImageProcessor ip) {
		width = ip.getWidth();
		height = ip.getHeight();
		Rectangle roi = ip.getRoi();
		if (roi != null) {
			roiX = roi.x;
			roiY = roi.y;
			roiWidth = roi.width;
			roiHeight = roi.height;
		}
		else {
			roiX = 0;
			roiY = 0;
			roiWidth = width;
			roiHeight = height;
		}
	}
	
	void getCentroid(ImageProcessor ip) {
		Rectangle roi = ip.getRoi();
		if (roi==null) {
			// no roi
			xCentroid = ip.getWidth()/2.0;
			yCentroid = ip.getHeight()/2.0;
			return;
		}
		int[] mask = ip.getMask();
		if (mask==null) {
			// rectangular roi
 			xCentroid = roi.x + roi.width/2.0;
			yCentroid = roi.y + roi.height/2.0;
			return;
		}
		// non-rectangular roi
		int count=0, xsum=0, ysum=0;
		for (int y=roiY,my=0; y<(roiY+roiHeight); y++,my++) {
			int mi = my*roiWidth;
			for (int x=roiX; x<(roiX+roiWidth); x++) {
				if (mask[mi++]==ip.BLACK)
					{count++; xsum+=x; ysum+=y;}
			}
		}
		xCentroid = (double)xsum/count+0.5;
		yCentroid = (double)ysum/count+0.5;
	}
	
}